-- name: Non-volatility
-- description: Allows some volatile objects to not be affected by the player so easily.\nCreated by Sunk.

gGlobalSyncTable.invincible = false
---@param obj Object
function goomba_loop(obj)
    if gGlobalSyncTable.invincible and obj.oAction == 102 then
        obj.oAction = 0
    end
end

---@param obj Object
function snufit_loop(obj)
    obj.oHealth = 0
    if gGlobalSyncTable.invincible then
        obj.oHealth = 1
    end
end

local do_not_replace = {
    [0] = true,
    [1] = true,
    [2] = true,
    [3] = true,
    [8] = true,
    [10] = true,
    [11] = true,
    [12] = true,
    [13] = true,
    [14] = true
}

---@param obj Object
function exclamation_box_loop(obj)
    if gGlobalSyncTable.invincible and obj.oAction == 6 and not do_not_replace[obj.oBehParams2ndByte] then
        obj.oAction = 5
        obj.oBehParams2ndByte = 15
    end
end

---@param obj Object
function crazy_box_init(obj)
    obj.oHomeX = obj.oPosX
    obj.oHomeY = obj.oPosY
    obj.oHomeZ = obj.oPosZ

    obj.oMoveAngleYaw = obj.oFaceAngleYaw
    obj.oMoveAnglePitch = obj.oFaceAnglePitch
    obj.oMoveAngleRoll = obj.oFaceAngleRoll
end


---@param obj Object
function object_unload(obj)
    if not gGlobalSyncTable.invincible then return end
    if cur_obj_has_behavior(get_behavior_from_id(id_bhvJumpingBox)) == 1 then
        spawn_non_sync_object(id_bhvJumpingBox, E_MODEL_BREAKABLE_BOX, obj.oHomeX, obj.oHomeY, obj.oHomeZ,
        ---@param o Object
        function (o)
            o.oFaceAngleYaw = obj.oMoveAngleYaw
            o.oFaceAnglePitch = obj.oMoveAnglePitch
            o.oFaceAngleRoll = obj.oMoveAngleRoll
        end)
    end
end

---@param m MarioState
function mario_update(m)
    if m.playerIndex ~= 0 then return end

    -- Thanks peachypeach
    if m.action ~= ACT_CRAZY_BOX_BOUNCE and m.heldObj ~= nil and obj_has_behavior_id(m.heldObj, id_bhvJumpingBox) == 1 then
        if m.action == ACT_HOLD_IDLE then
            set_mario_action(m, ACT_CRAZY_BOX_BOUNCE, 0)
        end
    end
end

---@param obj Object
function purple_switch_loop(obj)
    if not gGlobalSyncTable.invincible then return end
    if obj.oAction == PURPLE_SWITCH_TICKING then
        if obj.oTimer >= 150 then
            obj.oTimer = 0
            obj.oAction = PURPLE_SWITCH_UNPRESSED
        end
    end
end

local function switch(param, case_table)
    local case = case_table[param]
    if case then return case() end
    return nil
end

---@param obj Object
function cork_box_init(obj)
    obj.oFlags = OBJ_FLAG_UPDATE_GFX_POS_AND_ANGLE
    obj.collisionData = gGlobalObjectCollisionData.breakable_box_seg8_collision_08012D70
    obj.oCollisionDistance = 500
    bhv_init_room()

    obj.oHiddenObjectUnkF4 = nil
    obj.oAnimState = 1

    switch(obj.oBehParams2ndByte, {
        [0] = function ()
            obj.oNumLootCoins = 0
        end,
        [1] = function ()
            obj.oNumLootCoins = 3
        end,
        [2] = function ()
            obj.oNumLootCoins = 5
        end,
        [3] = function ()
            cur_obj_scale(1.5)
        end,
    })

    network_init_object(obj, true, nil)
end

--- @param parent Object
--- @param model ModelExtendedId
--- @param behavior BehaviorId
local function spawn_object(parent, model, behavior)
    if not parent then return nil end
    local obj = spawn_sync_object(behavior, model, parent.oPosX, parent.oPosY, parent.oPosZ,
        --- @param o Object
        function(o)
            -- Stars tend to break if they get assigned a parent object
            if o.behavior ~= get_behavior_from_id(id_bhvSpawnedStar) then
                o.parentObj = parent
            end
        end
    )
    if not obj then print("failed spawn"); return nil end

    obj_copy_pos_and_angle(obj, parent)

    return obj
end

---@param obj Object
---@param soundMagic any
local function create_sound_spawner(obj, soundMagic)
    local o = spawn_object(obj, E_MODEL_NONE, id_bhvSoundSpawner)
    if not o then return end

    o.oSoundEffectUnkF4 = soundMagic
end

---@param obj Object
function cork_box_loop(obj)
    load_object_collision_model()
    if not gGlobalSyncTable.invincible and (cur_obj_was_attacked_or_ground_pounded() ~= 0 or obj.oSyncDeath ~= 0) then
        obj_explode_and_spawn_coins(46, 1)
        create_sound_spawner(obj, SOUND_GENERAL_BREAK_BOX)
    end
end

hook_behavior(id_bhvGoomba, OBJ_LIST_GENACTOR, false, nil, goomba_loop, "bhvGoomba")
hook_behavior(id_bhvSnufit, OBJ_LIST_GENACTOR, false, nil, snufit_loop, "bhvSnufit")
hook_behavior(id_bhvExclamationBox, OBJ_LIST_SURFACE, false, nil, exclamation_box_loop, "bhvExclamationBox")
hook_behavior(id_bhvJumpingBox, OBJ_LIST_GENACTOR, false, crazy_box_init, nil)
hook_event(HOOK_ON_OBJECT_UNLOAD, object_unload)
hook_event(HOOK_MARIO_UPDATE, mario_update)
hook_behavior(id_bhvFloorSwitchGrills, OBJ_LIST_SURFACE, false, nil, purple_switch_loop)
hook_behavior(id_bhvBreakableBox, OBJ_LIST_SURFACE, true, cork_box_init, cork_box_loop)

if network_is_server() then
    hook_chat_command('invincible', "Makes some objects invincible. This either means they respawn or just cannot break or die.\nAffected objects are:\nGoombas, snufits, big breakable boxes, purple gate switches, exclamation boxes, and crazy boxes",
    function (msg)
        if msg == "off" then
            gGlobalSyncTable.invincible = false
        else
            gGlobalSyncTable.invincible = not gGlobalSyncTable.invincible
        end
        djui_chat_message_create("Affected objects are " .. (gGlobalSyncTable.invincible and "now" or "no longer") .. " invincible")
        return true
    end)
end